import pandas as pd
from sklearn.preprocessing import MinMaxScaler

from spylunking.log.setup_logging import build_colorized_logger
from antinex_utils.consts import SUCCESS
from antinex_utils.consts import ERR
from antinex_utils.consts import NOTRUN


log = build_colorized_logger(
    name='build_scaler_dataset')


def build_scaler_dataset_from_records(
        record_list,
        label="build-scaled-dataset",
        min_feature=-1,
        max_feature=1,
        cast_to_type="float32"):
    """build_scaler_dataset_from_records
    :param record_list: list of json records to scale between min/max
    :param label: log label for tracking
    :param min_feature: min feature range for scale normalization
    :param max_feature: max feature range for scale normalization
    :param cast_to_type: cast all of the dataframe to this datatype
    """

    status = NOTRUN
    last_step = "not-run"
    df = None
    scaler = None
    dataset = None

    res = {
        "status": status,
        "err": last_step,
        "org_recs": df,
        "scaler": scaler,
        "dataset": dataset
    }

    try:
        last_step = ("building scaler range=[{},{}]").format(
            min_feature,
            max_feature)
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        scaler = MinMaxScaler(
            feature_range=(
                min_feature,
                max_feature))

        last_step = ("converting records={} to df").format(
            len(record_list))
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        df = pd.read_json(record_list)
        if cast_to_type:
            last_step = ("casting df values to type={}").format(
                cast_to_type)
            log.info(("{} - {}")
                     .format(
                        label,
                        last_step))
            only_floats = df.values.astype(cast_to_type)

        last_step = ("running scale transform rows={}").format(
            len(df.index))
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        dataset = scaler.fit_transform(
            only_floats)

        status = SUCCESS
    except Exception as e:
        last_step = ("failed build_scaler_dataset_from_records "
                     "with ex={} last_step='{}' "
                     "recs={} range=[{},{}]").format(
                        e,
                        last_step,
                        str(record_list)[0:64],
                        min_feature,
                        max_feature)

        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        status = ERR
    # end of try/ex

    res = {
        "status": status,
        "err": last_step,
        "org_recs": df,
        "scaler": scaler,
        "dataset": dataset
    }
    return res
# end of build_scaler_dataset_from_records



import os
import uuid
import json
import numpy
import pandas as pd
import copy
from spylunking.log.setup_logging import build_colorized_logger
from antinex_utils.consts import SUCCESS
from antinex_utils.consts import ERR
from antinex_utils.consts import FAILED
from antinex_utils.consts import NOTRUN
from antinex_utils.utils import ev
from antinex_utils.utils import ppj
from antinex_utils.build_training_request import \
    build_training_request
from antinex_utils.build_scaler_dataset_from_records import \
    build_scaler_dataset_from_records
from antinex_utils.build_scaler_train_and_test_datasets import \
    build_scaler_train_and_test_datasets
from antinex_utils.merge_inverse_data_into_original import \
    merge_inverse_data_into_original
from keras.models import Sequential
from keras.models import model_from_json
from keras.layers import Dense
from keras.layers import Dropout
from keras.wrappers.scikit_learn import KerasRegressor
from keras.wrappers.scikit_learn import KerasClassifier
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split
import tensorflow as tf
import matplotlib
matplotlib.use("Agg")  # noqa
import matplotlib.pyplot as plt  # noqa


log = build_colorized_logger(
    name='make_predict')


def build_regression_dnn(
        num_features,
        compile_data,
        label="",
        model_json=None,
        model_desc=None):
    """build_regression_dnn

    :param num_features: input_dim for the number of
                         features in the data
    :param compile_data: dictionary of compile options
    :param label: log label for tracking this method
    :param model_json: keras model json to build the model
    :param model_desc: optional dictionary for model
    """

    model = Sequential()

    if model_json:
        log.info(("{} building regression "
                  "dnn model_json={}")
                 .format(
                     label,
                     model_json))
        model = model_from_json(json.dumps(model_json))
    elif model_desc:
        log.info(("{} building regression "
                  "dnn num_features={} model_desc={}")
                 .format(
                     label,
                     num_features,
                     model_desc))
        num_layers = 0
        for idx, node in enumerate(model_desc["layers"]):
            layer_type = node.get(
                "layer_type",
                "dense").lower()
            if layer_type == "dense":
                if num_layers == 0:
                    model.add(
                        Dense(
                            int(node["num_neurons"]),
                            input_dim=num_features,
                            kernel_initializer=node["init"],
                            activation=node["activation"]))
                else:
                    model.add(
                        Dense(
                            int(node["num_neurons"]),
                            kernel_initializer=node["init"],
                            activation=node["activation"]))
            else:
                if layer_type == "dropout":
                    model.add(
                        Dropout(
                            float(node["rate"])))
            # end of supported model types
            num_layers += 1
        # end of all layers
    else:
        # https://machinelearningmastery.com/regression-tutorial-keras-deep-learning-library-python/  # noqa
        log.info(("{} building regression "
                  "dnn num_features={}")
                 .format(
                     label,
                     num_features))
        model.add(
            Dense(
                8,
                input_dim=num_features,
                kernel_initializer="normal",
                activation="relu"))
        model.add(
            Dense(
                6,
                kernel_initializer="normal",
                activation="relu"))
        model.add(
            Dense(
                1,
                kernel_initializer="normal"))
    # end of building a regression dnn

    # if model was defined
    if model:
        log.info(("{} - regression compiling={}")
                 .format(
                    label,
                    compile_data))
        # compile the model
        loss = compile_data.get(
            "loss",
            "mse")
        optimizer = compile_data.get(
            "optimizer",
            "adam")
        metrics = compile_data.get(
            "metrics",
            [
                "mse",
                "mae",
                "mape",
                "cosine"
            ])
        model.compile(
                loss=loss,
                optimizer=optimizer,
                metrics=metrics)
    else:
        log.error(("{} - failed building regression model")
                  .format(
                      label))
    # if could compile model

    return model
# end of build_regression_dnn


def build_classification_dnn(
        num_features,
        compile_data,
        label="",
        model_json=None,
        model_desc=None):
    """build_classification_dnn

    :param num_features: input_dim for the number of
                         features in the data
    :param compile_data: dictionary of compile options
    :param label: log label for tracking this method
    :param model_json: keras model json to build the model
    :param model_desc: optional dictionary for model
    """

    model = Sequential()

    if model_json:
        log.info(("{} building classification "
                  "dnn model_json={}")
                 .format(
                     label,
                     model_json))
        model = model_from_json(json.dumps(model_json))
    elif model_desc:
        log.info(("{} building classification "
                  "dnn num_features={} model_desc={}")
                 .format(
                     label,
                     num_features,
                     model_desc))
        num_layers = 0
        for idx, node in enumerate(model_desc["layers"]):
            layer_type = node.get(
                "layer_type",
                "dense").lower()
            if layer_type == "dense":
                if num_layers == 0:
                    model.add(
                        Dense(
                            int(node["num_neurons"]),
                            input_dim=num_features,
                            kernel_initializer=node["init"],
                            activation=node["activation"]))
                else:
                    model.add(
                        Dense(
                            int(node["num_neurons"]),
                            kernel_initializer=node["init"],
                            activation=node["activation"]))
            else:
                if layer_type == "dropout":
                    model.add(
                        Dropout(
                            float(node["rate"])))
            # end of supported model types
            num_layers += 1
        # end of all layers
    else:
        # https://machinelearningmastery.com/tutorial-first-neural-network-python-keras/  # noqa
        log.info(("{} building classification "
                  "dnn num_features={}")
                 .format(
                     label,
                     num_features))
        model.add(
            Dense(
                8,
                input_dim=num_features,
                kernel_initializer="uniform",
                activation="relu"))
        model.add(
            Dense(
                6,
                kernel_initializer="uniform",
                activation="relu"))
        model.add(
            Dense(
                1,
                kernel_initializer="uniform",
                activation="sigmoid"))
    # end of building a classification dnn

    # if model was defined
    if model:
        log.info(("{} - classification compiling={}")
                 .format(
                    label,
                    compile_data))
        # compile the model
        loss = compile_data.get(
            "loss",
            "binary_crossentropy")
        optimizer = compile_data.get(
            "optimizer",
            "adam")
        metrics = compile_data.get(
            "metrics",
            [
                "accuracy"
            ])
        model.compile(
                loss=loss,
                optimizer=optimizer,
                metrics=metrics)
    else:
        log.error(("{} - failed building classification model")
                  .format(
                      label))
    # if could compile model

    return model
# end of build_classification_dnn


def check_request(
        req):
    """check_request

    :param req: dictionary to check values
    """
    label = req.get("label", "no-label-set")
    predict_rows = req.get("predict_rows", None)
    manifest = req.get("manifest", None)
    dataset = req.get("dataset", None)
    csv_file = req.get("csv_file", None)
    use_existing_model = req.get("use_existing_model", False)

    if use_existing_model:
        if not predict_rows and not dataset:
            return ("{} missing predict_rows or dataset for existing model"
                    "request={}").format(
                        label,
                        req)
        else:
            return None
    # existing models just need a couple rows or a dataset to work

    if not manifest and not dataset and not predict_rows and not csv_file:
        return ("{} missing manifest "
                "request={}").format(
                    label,
                    ppj(req))
    if manifest:
        csv_file = manifest.get(
            "csv_file",
            None)
        if not predict_rows and not csv_file and not dataset:
            return ("{} missing dataset predict_rows or csv_file in "
                    "manifest of request={}").format(
                        label,
                        ppj(req))

    if not dataset and not predict_rows and not csv_file:
        return ("{} missing dataset or predict_rows or csv_file in "
                "request={}").format(
                    label,
                    ppj(req))

    return None
# end of check_request


def save_prediction_image(
        label="not-set",
        history=None,
        histories=[],
        image_file=None):
    """save_prediction_image

    :param history: model prediction history
    :param histories: histories to generate in the image
    :param image_file: save to file
    """

    status = FAILED

    if not history:
        log.info(("{} - no history")
                 .format(
                     label))
        return status
    if not histories:
        log.info(("{} - no histories")
                 .format(
                     label))
        return status
    if not image_file:
        log.info(("{} - no image_file")
                 .format(
                     label))
        return status

    try:
        if history and len(histories) > 0:
            log.info(("plotting history={} "
                      "histories={}")
                     .format(
                        history,
                        histories))
            should_save = False
            for h in histories:
                if h in history.history:
                    log.info(("plotting={}")
                             .format(
                                h))
                    plt.plot(
                        history.history[h],
                        label=h)
                    should_save = True
                else:
                    log.error(("missing history={}")
                              .format(
                                h))
            # for all histories

            if should_save:
                log.info(("saving plots as image={}")
                         .format(
                            image_file))
                plt.legend(loc='best')
                plt.savefig(image_file)
                if not os.path.exists(image_file):
                    log.error(("Failed saving image={}")
                              .format(
                                    image_file))
            # end of saving file

        # end of if there are histories to plot
    except Exception as e:
        log.error(("Failed saving "
                   "image_file={} ex={}")
                  .format(
                    image_file,
                    e))
    # end of try/ex
# end of save_prediction_image


def build_train_and_test_features(
        df_columns,
        features_to_process,
        predict_feature,
        ignore_features):
    """build_train_and_test_features

    Order matters when slicing up datasets using scalers...

    if not, then something that is an int/bool can get into a float
    column and that is really bad for making predictions
    with new or pre-trained models...

    :param df_columns: columns in the dataframe
    :param features_to_process: requested features to train
    :param predict_feature: requested feature to predict
    :param ignore_features: requested non-numeric/not-wanted features
    """
    train_and_test_features = []

    # for all columns in the data
    # add columns in order if they are in the requested:
    # features_to_process list and not in the ignore_features
    for c in df_columns:
        if c == predict_feature:
            train_and_test_features.append(
                c)
        else:
            add_feature = True
            for i in ignore_features:
                if i == c:
                    add_feature = False
                    break
            if add_feature:
                for f in features_to_process:
                    if f == c:
                        train_and_test_features.append(
                            c)
                        break
    # end of filtering features before scalers

    return train_and_test_features
# end of build_train_and_test_features


def make_predictions(
        req):
    """make_predictions

    :param req: dictionary for making predictions
    """

    last_step = "not-run"
    label = "no-label-set"
    model = None
    predictions = None
    sample_predictions = []
    rounded = []
    accuracy = None
    error = None
    image_file = None
    history = None
    histories = None
    indexes = None
    scores = None
    cm = None
    scaler_res = None
    scaler_res_data = None
    scaler_train = None
    scaler_test = None
    scaled_train_dataset = None
    scaled_test_dataset = None
    inverse_predictions = None
    merge_df = None
    are_predicts_merged = False
    existing_model_dict = None
    data = {
        "predictions": predictions,
        "rounded_predictions": rounded,
        "sample_predictions": sample_predictions,
        "acc": accuracy,
        "scores": scores,
        "history": history,
        "histories": histories,
        "image_file": image_file,
        "model": model,
        "indexes": indexes,
        "confusion_matrix": cm,
        "scaler_train": scaler_train,
        "scaler_test": scaler_test,
        "scaled_train_dataset": scaled_train_dataset,
        "scaled_test_dataset": scaled_test_dataset,
        "inverse_predictions": inverse_predictions,
        "apply_scaler": False,
        "are_predicts_merged": are_predicts_merged,
        "merge_df": merge_df,
        "err": error
    }
    res = {
        "status": NOTRUN,
        "err": last_step,
        "data": None
    }

    try:

        label = req.get("label", "no-label-set")
        last_step = "validating"
        log.info("{} - {}".format(
            label,
            last_step))

        invalid_error_string = check_request(req)
        if invalid_error_string:
            last_step = ("{} - {}").format(
                            label,
                            invalid_error_string)
            log.info(("predictions stopping: {}")
                     .format(
                        last_step))
            res["err"] = last_step
            res["status"] = ERR
            res["data"] = None
            return res
        # end of checking for bad request inputs

        predict_rows = req.get("predict_rows", None)
        verbose = int(req.get("verbose", "1"))
        manifest = req.get("manifest", {})
        model_json = req.get("model_json", None)
        model_desc = req.get("model_desc", None)
        weights_json = req.get("weights_json", None)
        weights_file = req.get("weights_file", None)
        should_predict = req.get("should_predict", True)
        dataset = req.get("dataset", None)
        new_model = True
        existing_model_dict = req.get("use_existing_model", None)
        if existing_model_dict:
            log.info(("{} - using existing model={}")
                     .format(
                        label,
                        existing_model_dict["model"]))
            accuracy = existing_model_dict["acc"]
            scores = existing_model_dict["scores"]
            history = existing_model_dict["history"]
            histories = existing_model_dict["histories"]
            model = existing_model_dict["model"]
            new_model = False
        else:
            log.info(("{} - new model")
                     .format(
                        label))
        # end of loading the existing model

        save_weights = False
        image_file = req.get("image_file", None)
        loss = req.get(
            "loss",
            manifest.get(
                "loss",
                "mse"))
        optimizer = req.get(
            "optimizer",
            manifest.get(
                "optimizer",
                "adam"))
        metrics = req.get(
            "metrics",
            manifest.get(
                "metrics",
                [
                    "accuracy"
                ]))
        histories = req.get(
            "histories",
            manifest.get(
                "histories",
                [
                    "val_loss",
                    "val_acc",
                    "loss",
                    "acc"
                ]))
        ml_type = req.get(
            "ml_type",
            manifest.get(
                "ml_type",
                "classification"))
        features_to_process = req.get(
            "features_to_process",
            manifest.get(
                "features_to_process",
                []))
        filter_features = req.get(
            "filter_features",
            [])
        ignore_features = req.get(
            "ignore_features",
            [])
        predict_feature = req.get(
            "predict_feature",
            manifest.get(
                "predict_feature",
                None))
        epochs = int(req.get(
            "epochs",
            manifest.get(
                "epochs",
                "5")))
        batch_size = int(req.get(
            "batch_size",
            manifest.get(
                "batch_size",
                "32")))
        test_size = float(req.get(
            "test_size",
            manifest.get(
                "test_size",
                "0.2")))
        num_splits = int(req.get(
            "num_splits",
            manifest.get(
                "num_splits",
                "2")))
        verbose = int(req.get(
            "verbose",
            manifest.get(
                "verbose",
                "1")))
        seed = int(req.get(
            "seed",
            manifest.get(
                "seed",
                "9")))
        label_rules = req.get(
            "label_rules",
            manifest.get(
                "label_rules",
                {}))
        min_scaler_range = int(req.get(
            "min_scaler_range",
            "-1"))
        max_scaler_range = int(req.get(
            "max_scaler_range",
            "1"))
        apply_scaler = bool(str(req.get(
            "apply_scaler",
            "false")).lower() == "true")
        scaler_cast_to_type = req.get(
            "scaler_cast_type",
            "float32")
        sort_by = req.get(
            "sort_values",
            None)
        max_records = int(req.get(
            "max_records",
            "100000"))
        predict_type = manifest.get(
            "predict_type",
            "predict")
        manifest_headers = manifest.get(
            "headers",
            [])
        csv_file = manifest.get(
            "csv_file",
            None)
        meta_file = manifest.get(
            "meta_file",
            None)
        if not weights_file:
            weights_file = manifest.get(
                "model_weights_file",
                None)
        num_features = len(features_to_process)
        detected_headers = []
        num_samples = None
        org_df = None
        row_df = None
        filter_df = None
        sample_rows = None
        target_rows = None
        num_target_rows = None
        ml_req = None
        use_evaluate = False
        if csv_file and meta_file and predict_feature:
            if os.path.exists(csv_file) and os.path.exists(meta_file):
                use_evaluate = True
        else:
            if dataset:
                if os.path.exists(dataset):
                    use_evaluate = True
                    csv_file = dataset
            else:
                if predict_rows and not existing_model_dict:
                    use_evaluate = True
        # end of if we're building a dataset from these locations

        numpy.random.seed(seed)

        last_step = ("loading prediction "
                     "into dataframe seed={} "
                     "scaler={} range[{},{}]").format(
                        seed,
                        apply_scaler,
                        min_scaler_range,
                        max_scaler_range)
        log.info("{} - {}".format(
            label,
            last_step))
        if not weights_file:
            weights_file = manifest.get(
                "model_weights_file",
                None)

        last_step = ("loading prediction into dataframe")

        log.info("{} - {}".format(
            label,
            last_step))

        # convert json into pandas dataframe for model.predict
        try:
            if new_model and use_evaluate and not predict_rows:
                log.info(("{} - loading predictions new_model={} "
                          "evaluate={} csv={} sort={}")
                         .format(
                            label,
                            new_model,
                            use_evaluate,
                            csv_file,
                            sort_by))
                if sort_by:
                    org_df = pd.read_csv(
                            csv_file,
                            encoding="utf-8-sig").sort_values(
                            by=sort_by)
                else:
                    org_df = pd.read_csv(
                            csv_file,
                            encoding="utf-8-sig")

                predict_rows = org_df.to_json()
                detected_headers = list(org_df.columns.values)

                if apply_scaler:

                    for f in features_to_process:
                        if f not in org_df.columns:
                            log.error(("{} "
                                       "csv={} is missing column={}")
                                      .format(
                                        label,
                                        csv_file,
                                        f))
                    # show columns that were supposed to be in the
                    # dataset but are not

                    train_and_test_features = \
                        build_train_and_test_features(
                            org_df.columns,
                            features_to_process,
                            predict_feature,
                            ignore_features)

                    log.info(("building csv scalers all_features={}")
                             .format(
                                train_and_test_features))

                    scaler_transform_res = \
                        build_scaler_dataset_from_records(
                            label=label,
                            record_list=org_df[
                                train_and_test_features].to_json(),
                            min_feature=min_scaler_range,
                            max_feature=max_scaler_range,
                            cast_to_type=scaler_cast_to_type)

                    if scaler_transform_res["status"] == SUCCESS:
                        log.info(("{} - scaled dataset predict_rows={} "
                                  "df={} dataset={}")
                                 .format(
                            label,
                            len(predict_rows),
                            len(scaler_transform_res["org_recs"].index),
                            len(scaler_transform_res["dataset"])))
                    else:
                        log.error(("{} - failed to scale dataset err={} "
                                   "predict_rows={} df={} dataset={}")
                                  .format(
                            label,
                            scaler_transform_res["err"],
                            len(predict_rows),
                            len(scaler_transform_res["org_recs"].index),
                            len(scaler_transform_res["dataset"])))
                    # if scaler works on dataset

                    log.info(("{} building scaled samples and rows")
                             .format(
                                label))
                    # noqa https://stackoverflow.com/questions/21764475/scaling-numbers-column-by-column-with-pandas-python
                    row_df = pd.DataFrame(
                                scaler_transform_res["dataset"],
                                columns=org_df[
                                    train_and_test_features].columns)
                    sample_rows = row_df[features_to_process]
                    target_rows = row_df[predict_feature]
                    num_samples = len(sample_rows.index)
                    num_target_rows = len(target_rows.index)
                else:
                    log.info(("{} - not applying scaler to predict_rows")
                             .format(
                                label))
                    row_df = org_df
                    log.info(("{} - setting samples "
                              "to features_to_process={} cols={}")
                             .format(
                                label,
                                ppj(features_to_process),
                                list(org_df.columns.values)))
                    sample_rows = row_df[features_to_process]
                    target_rows = row_df[predict_feature]
                    num_samples = len(sample_rows.index)
                    num_target_rows = len(target_rows.index)
                # if applying scaler to predict rows
            # end of loading from a csv

            if dataset:
                log.info(("{} loading dataset={}")
                         .format(
                            label,
                            dataset))

                if sort_by:
                    org_df = pd.read_csv(
                        dataset,
                        encoding="utf-8-sig").sort_values(
                            by=sort_by)
                else:
                    org_df = pd.read_csv(
                        dataset,
                        encoding="utf-8-sig")
                # end of loading the df

                log.info(("{} preparing dataset={}")
                         .format(
                            label,
                            dataset))
                ml_req = {
                    "X_train": None,
                    "Y_train": None,
                    "X_test": None,
                    "Y_test": None
                }
                scaled_train_features = []
                cur_headers = list(org_df.columns.values)
                for h in cur_headers:
                    include_feature = True
                    if h == predict_feature:
                        include_feature = False
                    else:
                        for f in features_to_process:
                            if h == f:
                                scaled_train_features.append(h)
                                include_feature = False
                                break
                        for e in ignore_features:
                            if h == e:
                                include_feature = False
                                break
                    # filter out columns

                    if include_feature:
                        features_to_process.append(h)
                # end of building features

                # make sure to prune out ignored ones:
                cleaned_scaled_train = []
                for idx, h in enumerate(scaled_train_features):
                    should_include = True
                    for i in ignore_features:
                        if h == i:
                            should_include = False
                    if should_include:
                        cleaned_scaled_train.append(h)
                # end of pruning ignored ones

                # assign to cleaned list
                scaled_train_features = cleaned_scaled_train

                filter_features = copy.deepcopy(features_to_process)
                include_predict_feature = True
                for f in filter_features:
                    if f == predict_feature:
                        include_predict_feature = False
                        break

                if include_predict_feature:
                    filter_features.append(predict_feature)

                if apply_scaler:

                    num_features = len(scaled_train_features)
                    log.info(("{} scaling dataset={} "
                              "scaled_train_features={}")
                             .format(
                                label,
                                len(org_df.index),
                                ppj(scaled_train_features)))

                    scaler_res = \
                        build_scaler_train_and_test_datasets(
                            label=label,
                            train_features=scaled_train_features,
                            test_feature=predict_feature,
                            df=org_df,
                            test_size=test_size,
                            seed=seed,
                            scaler_cast_to_type=scaler_cast_to_type,
                            min_feature_range=min_scaler_range,
                            max_feature_range=max_scaler_range)

                    if scaler_res["status"] != SUCCESS:
                        log.info(("{} - scaler transform failed error={}")
                                 .format(
                                    label,
                                    scaler_res["err"]))
                        res["status"] = ERR
                        res["err"] = last_step
                        res["data"] = None
                        return res
                    else:
                        log.info(("{} - scaler transform done")
                                 .format(
                                    label))
                        scaler_res_data = scaler_res["data"]
                        ml_req["X_train"] = scaler_res_data["x_train"]
                        ml_req["Y_train"] = scaler_res_data["y_train"]
                        ml_req["X_test"] = scaler_res_data["x_test"]
                        ml_req["Y_test"] = scaler_res_data["y_test"]

                        scaler_train = scaler_res_data["scaler_train"]
                        scaler_test = scaler_res_data["scaler_test"]
                        scaled_train_dataset = \
                            scaler_res_data["scaled_train_dataset"]
                        scaled_test_dataset = \
                            scaler_res_data["scaled_test_dataset"]

                        log.info(("{} - building scaled row_df "
                                  "filter_features={}")
                                 .format(
                                    label,
                                    filter_features))

                        # noqa https://stackoverflow.com/questions/21764475/scaling-numbers-column-by-column-with-pandas-python
                        last_step = ("building row_df from scaled ds "
                                     "train_features={}").format(
                                        scaled_train_features)
                        row_df = pd.DataFrame(
                            scaler_res_data["scaled_train_dataset"],
                            columns=list(scaled_train_features))
                        last_step = ("building samples from rows_df={} "
                                     "train_features={}").format(
                                        len(row_df.index),
                                        scaled_train_features)
                        sample_rows = pd.DataFrame(
                            scaler_res_data["scaled_train_dataset"],
                            columns=list(scaled_train_features))
                        last_step = ("building targets from scaled ds "
                                     "predict_feature={}").format(
                                        predict_feature)
                        target_rows = pd.DataFrame(
                            scaler_res_data["scaled_test_dataset"],
                            columns=[predict_feature])
                        last_step = ("adding predict_feature={} to "
                                     "row_df").format(
                                        predict_feature)
                        row_df[predict_feature] = \
                            scaler_res_data["scaled_test_dataset"]
                        last_step = ("counting num_sample_rows")
                        num_samples = len(sample_rows.index)
                        last_step = ("counting num_target_rows")
                        num_target_rows = len(target_rows.index)
                        log.info(("{} row_df created features={} "
                                  "num_samples={} num_targets={}")
                                 .format(
                                    label,
                                    features_to_process,
                                    num_samples,
                                    num_target_rows))
                    # end of setting up scaler train/test data
                else:
                    num_features = len(features_to_process)
                    log.info(("{} filtering dataset={} filter_features={}")
                             .format(
                                label,
                                len(org_df.index),
                                ppj(filter_features)))
                    filter_df = org_df[filter_features]

                    log.info(("{} splitting non-scaled"
                              "filtered_df={} predict_feature={} test_size={} "
                              "features={} ignore_features={} csv={}")
                             .format(
                                label,
                                len(filter_df.index),
                                test_size,
                                predict_feature,
                                ppj(features_to_process),
                                ppj(ignore_features),
                                dataset))
                    # split the data into training

                    (ml_req["X_train"],
                     ml_req["X_test"],
                     ml_req["Y_train"],
                     ml_req["Y_test"]) = train_test_split(
                        filter_df[features_to_process],
                        filter_df[predict_feature],
                        test_size=test_size,
                        random_state=seed)

                    row_df = org_df
                    log.info(("{} - setting samples "
                              "to features_to_process={} cols={}")
                             .format(
                                label,
                                ppj(features_to_process),
                                list(org_df.columns.values)))
                    sample_rows = row_df[features_to_process]
                    target_rows = row_df[predict_feature]
                    num_samples = len(sample_rows.index)
                    num_target_rows = len(target_rows.index)
                # if applying scaler to predict rows
            else:
                if apply_scaler:
                    log.info(("{} - no dataset - scaling predict_rows={}")
                             .format(
                                label,
                                len(predict_rows)))

                    org_df = pd.read_json(predict_rows)
                    row_df = org_df

                    for f in features_to_process:
                        if f not in org_df.columns:
                            log.error(("{} "
                                       "predict_rows are missing column={}")
                                      .format(
                                        label,
                                        f))
                    # show columns that were supposed to be in the
                    # predict_rows but are not

                    train_and_test_features = \
                        build_train_and_test_features(
                            org_df.columns,
                            features_to_process,
                            predict_feature,
                            ignore_features)

                    log.info(("building predict_rows scalers all_features={}")
                             .format(
                                train_and_test_features))

                    scaler_transform_res = \
                        build_scaler_dataset_from_records(
                            label=label,
                            record_list=org_df[
                                train_and_test_features].to_json(),
                            min_feature=min_scaler_range,
                            max_feature=max_scaler_range,
                            cast_to_type=scaler_cast_to_type)

                    if scaler_transform_res["status"] == SUCCESS:
                        log.info(("{} - scaled predict_rows={} "
                                  "df={} dataset={}")
                                 .format(
                            label,
                            len(predict_rows),
                            len(scaler_transform_res["org_recs"].index),
                            len(scaler_transform_res["dataset"])))
                    else:
                        log.error(("{} - failed to scale predict err={} "
                                   "predict_rows={} df={} dataset={}")
                                  .format(
                            label,
                            scaler_transform_res["err"],
                            len(predict_rows),
                            len(scaler_transform_res["org_recs"].index),
                            len(scaler_transform_res["dataset"])))
                    # if scaler works on dataset

                    log.info(("{} building predict org_df scaled "
                              "for testing all samples and predict_rows")
                             .format(
                                label))

                    # noqa https://stackoverflow.com/questions/21764475/scaling-numbers-column-by-column-with-pandas-python
                    row_df = pd.DataFrame(
                                scaler_transform_res["dataset"],
                                columns=org_df[
                                    train_and_test_features].columns)

                    log.info(("{} casting data to floats")
                             .format(
                                label))

                    ml_req = {
                        "X_train": row_df[features_to_process].astype(
                            "float32").values,
                        "Y_train": row_df[predict_feature].astype(
                            "float32").values,
                        "X_test": row_df[features_to_process].astype(
                            "float32").values,
                        "Y_test": row_df[predict_feature].astype(
                            "float32").values
                    }

                    sample_rows = row_df[features_to_process]
                    target_rows = row_df[predict_feature]
                    num_samples = len(sample_rows.index)
                    num_target_rows = len(target_rows.index)
                else:
                    log.info(("{} - no dataset using org_df")
                             .format(
                                label))

                    org_df = pd.read_json(predict_rows)
                    row_df = org_df

                    ml_req = {
                        "X_train": org_df[features_to_process].astype(
                            "float32").values,
                        "Y_train": org_df[predict_feature].astype(
                            "float32").values,
                        "X_test": org_df[features_to_process].astype(
                            "float32").values,
                        "Y_test": org_df[predict_feature].astype(
                            "float32").values
                    }

                    log.info(("{} building predict org_df scaled "
                              "for testing all samples WITHOUT predict_rows")
                             .format(
                                label))

                    log.info(("{} - setting samples "
                              "to features_to_process={} cols={}")
                             .format(
                                label,
                                ppj(features_to_process),
                                list(org_df.columns.values)))
                    sample_rows = row_df[features_to_process]
                    target_rows = row_df[predict_feature]
                    num_samples = len(sample_rows.index)
                    num_target_rows = len(target_rows.index)
                # end of if apply_scalar to csv + predict_rows
            # end of handling metadata-driven split vs controlled

        except Exception as f:
            last_step = ("{} - failed during '{}' json={} "
                         "with ex={}").format(
                            label,
                            last_step,
                            model_json,
                            f)
            log.error(last_step)
            res["status"] = ERR
            res["err"] = last_step
            res["data"] = None
            return res
        # end of try/ex to convert rows to pandas dataframe

        if num_samples == 0:
            last_step = ("{} - "
                         "missing predict_rows={}").format(
                            label,
                            sample_rows)
            log.error(last_step)
            res["status"] = ERR
            res["err"] = last_step
            res["data"] = None
            return res
        # stop if no rows

        # check the headers in the dataframe match what
        # the original model was trained with
        for h in manifest_headers:
            found_header = False
            for d in detected_headers:
                if h == d:
                    found_header = True
                    break
            # end of for all detected headers
            if not found_header:
                last_step = ("{} - invalid predict_rows - header={} in "
                             "detected_headers={}").format(
                        label,
                        h,
                        detected_headers)
                log.error(last_step)
                res["status"] = ERR
                res["err"] = last_step
                res["data"] = None
                return res
        # end for all manifest headers - expected to be in
        # predict_rows

        # create a back up weights file if one was not targeted
        if not weights_file:
            h5_storage_dir = ev(
                "H5_DIR",
                "/tmp")
            weights_file = "{}/{}.h5".format(
                h5_storage_dir,
                str(uuid.uuid4()))
        # end of building a weights file

        last_step = "loading model"
        log.info("{} - {}".format(
            label,
            last_step))

        # load the model from the json
        try:

            if new_model:
                compile_data = {
                    "loss": loss,
                    "optimizer": optimizer,
                    "metrics": metrics
                }

                if ml_type == "standalone-classification":
                    model = build_classification_dnn(
                            num_features=num_features,
                            compile_data=compile_data,
                            label=label,
                            model_json=model_json,
                            model_desc=model_desc)
                elif ml_type == "classification":
                    def set_model():
                        return build_classification_dnn(
                            num_features=num_features,
                            compile_data=compile_data,
                            label=label,
                            model_json=model_json,
                            model_desc=model_desc)

                    model = KerasClassifier(
                            build_fn=set_model,
                            epochs=epochs,
                            batch_size=batch_size,
                            verbose=verbose)
                elif ml_type == "regression":
                    def set_model():
                        return build_regression_dnn(
                            num_features=num_features,
                            compile_data=compile_data,
                            label=label,
                            model_json=model_json,
                            model_desc=model_desc)

                    model = KerasRegressor(
                            build_fn=set_model,
                            epochs=epochs,
                            batch_size=batch_size,
                            verbose=verbose)
                else:
                    def set_model():
                        return build_regression_dnn(
                            num_features=num_features,
                            compile_data=compile_data,
                            label=label,
                            model_json=model_json,
                            model_desc=model_desc)

                    model = KerasRegressor(
                            build_fn=set_model,
                            epochs=epochs,
                            batch_size=batch_size,
                            verbose=verbose)
            else:
                log.info(("{} - using existing - not building={}")
                         .format(
                            label,
                            new_model))
            # end of if new_model or use existing
        except Exception as f:
            last_step = ("{} - failed during '{}' ml_type={} "
                         "model_json={} model_desc={} "
                         "with ex={}").format(
                            label,
                            last_step,
                            ml_type,
                            model_json,
                            model_desc,
                            f)
            log.error(last_step)
            res["status"] = ERR
            res["err"] = last_step
            res["data"] = None
            return res
        # end of try/ex to save weights

        last_step = "saving model weights={}".format(
            weights_file)
        log.info("{} - {}".format(
            label,
            last_step))

        if use_evaluate:
            last_step = ("building training ml_type={} "
                         "predict_type={} "
                         "sample_rows={} target_rows={} "
                         "manifest={}").format(
                            ml_type,
                            predict_type,
                            num_samples,
                            num_target_rows,
                            ppj(manifest))
            log.info("{} - {}".format(
                label,
                last_step))

            # build training request for new predicts

            if not dataset and csv_file:
                ml_req = build_training_request(
                        csv_file=csv_file,
                        meta_file=meta_file,
                        predict_feature=predict_feature,
                        test_size=test_size)

            # fit the model
            if new_model:
                last_step = ("fitting Xtrain={} Ytrain={} Xtest={} Ytest={} "
                             "epochs={} batch_size={}").format(
                                len(ml_req["X_train"]),
                                len(ml_req["Y_train"]),
                                len(ml_req["X_test"]),
                                len(ml_req["Y_test"]),
                                epochs,
                                batch_size)
                log.info("{} - {}".format(
                    label,
                    last_step))
                history = model.fit(
                        ml_req["X_train"],
                        ml_req["Y_train"],
                        validation_data=(
                            ml_req["X_test"],
                            ml_req["Y_test"]),
                        epochs=epochs,
                        batch_size=batch_size,
                        shuffle=False,
                        verbose=verbose)
            else:
                log.info(("{} - using existing - not fitting={}")
                         .format(
                            label,
                            new_model))
            # end of if new model or using existing

        # end of building training update data
        # not use_evaluate

        last_step = ("predicting ml_type={} predict_type={} "
                     "rows={} manifest={}").format(
                        ml_type,
                        predict_type,
                        num_samples,
                        ppj(manifest))
        log.info("{} - {}".format(
            label,
            last_step))

        if os.path.exists(weights_file):
            last_step = ("loading weights_file={}").format(
                            weights_file)
            log.info("{} - {}".format(
                label,
                last_step))

            # load the weights from the file on disk
            try:
                if ml_type == "standalone-classification":
                    model.load_weights(
                        weights_file)
                else:
                    model.model.load_weights(
                        weights_file)
            except Exception as f:
                last_step = ("{} - failed during '{}' "
                             "file={} weights={}"
                             "with ex={}").format(
                                label,
                                last_step,
                                weights_file,
                                weights_json,
                                f)
                log.error(last_step)
                res["status"] = ERR
                res["err"] = last_step
                res["data"] = None
                return res
            # end of try/ex to save weights
        else:
            log.info(("{} did not find weights_file={}")
                     .format(
                        label,
                        weights_file))
        # only load weights if the file is still on disk

        # evaluating
        last_step = ("evaluating num_xtest={} num_ytest={} "
                     "metrics={} histories={} "
                     "loss={} optimizer={}").format(
                        len(ml_req["X_test"]),
                        len(ml_req["Y_test"]),
                        metrics,
                        histories,
                        loss,
                        optimizer)
        log.info("{} - {}".format(
            label,
            last_step))

        # make predictions
        try:
            if should_predict:
                if ml_type == "standalone-classification":
                    if new_model:
                        scores = model.evaluate(
                            ml_req["X_test"],
                            ml_req["Y_test"])
                        if len(scores) > 1:
                            accuracy = {
                                "accuracy": scores[1] * 100
                            }
                        else:
                            accuracy = {
                                "accuracy": 0.0
                            }
                    # no scoring on existing models
                    predictions = model.predict(
                        sample_rows.values,
                        verbose=verbose)
                    numpy.set_printoptions(threshold=numpy.nan)
                    sess = tf.InteractiveSession()  # noqa
                    indexes = tf.argmax(predictions, axis=1)
                    data["indexes"] = indexes
                    rounded = [round(x[0]) for x in predictions]
                    ridx = 0
                    should_set_labels = False
                    labels_dict = {}
                    if "labels" in label_rules \
                       and "label_values" in label_rules:
                        label_rows = label_rules["label_values"]
                        for idx, lidx in enumerate(label_rows):
                            if len(label_rules["labels"]) >= idx:
                                should_set_labels = True
                                labels_dict[str(lidx)] = \
                                    label_rules["labels"][idx]
                    # end of compiling labels dictionary
                    log.info(("{} - scores={} accuracy={} "
                              "merging predictions={} labels={}")
                             .format(
                                label,
                                scores,
                                accuracy.get("accuracy", None),
                                len(sample_rows.index),
                                labels_dict))
                    for idx, row in row_df.iterrows():
                        if len(sample_predictions) > max_records:
                            log.info(("{} hit max={} predictions")
                                     .format(
                                         label,
                                         max_records))
                            break
                        new_row = json.loads(row.to_json())
                        cur_value = rounded[ridx]
                        if predict_feature in row:
                            new_row["_original_{}".format(
                                    predict_feature)] = \
                                row[predict_feature]
                        else:
                            new_row["_original_{}".format(
                                    predict_feature)] = \
                                "missing-from-dataset"
                        new_row[predict_feature] = int(cur_value)
                        if should_set_labels:
                            new_row["label_name"] = \
                                labels_dict[str(int(cur_value))]
                        new_row["_row_idx"] = ridx
                        new_row["_count"] = idx
                        sample_predictions.append(new_row)
                        ridx += 1
                    # end of merging samples with predictions
                elif ml_type == "classification":
                    if new_model:
                        last_step = "building estimators"
                        estimators = []
                        estimators.append(
                            ("standardize",
                             StandardScaler()))
                        estimators.append(
                            ("mlp",
                             model))
                        last_step = "building pipeline"
                        pipeline = Pipeline(estimators)
                        # https://machinelearningmastery.com/multi-class-classification-tutorial-keras-deep-learning-library/  # noqa
                        last_step = ("{} - starting classification "
                                     "StratifiedKFold "
                                     "splits={} seed={}").format(
                                        label,
                                        num_splits,
                                        seed)
                        log.info(last_step)
                        kfold = StratifiedKFold(
                            n_splits=num_splits,
                            random_state=seed)
                        last_step = "cross_val_score"
                        log.info(("{} - classification cross_val_score: ")
                                 .format(
                                    label))
                        results = cross_val_score(
                            pipeline,
                            ml_req["X_train"],
                            ml_req["Y_train"],
                            cv=kfold)
                        scores = [
                            results.std(),
                            results.mean()
                        ]
                        accuracy = {
                            "accuracy": results.mean() * 100
                        }
                        log.info(("{} - classification accuracy={} samples={}")
                                 .format(
                                    label,
                                    accuracy["accuracy"],
                                    num_samples))
                    else:
                        log.info(("{} - using existing "
                                  "accuracy={} scores={} "
                                  "predictions={}")
                                 .format(
                                    label,
                                    accuracy,
                                    scores,
                                    len(sample_rows.index)))
                    # end of if use existing or new model
                    predictions = model.predict(
                        sample_rows.values,
                        verbose=verbose)
                    if new_model:
                        log.info(("{} - "
                                  "classification confusion_matrix samples={} "
                                  "predictions={} target_rows={}")
                                 .format(
                                    label,
                                    num_samples,
                                    len(predictions),
                                    num_target_rows))
                        cm = confusion_matrix(
                            target_rows.values,
                            predictions)
                        log.info(("{} - "
                                  "classification has confusion_matrix={} "
                                  "predictions={} target_rows={}")
                                 .format(
                                    label,
                                    cm,
                                    len(predictions),
                                    num_target_rows))
                    # end of confusion matrix
                    rounded = [round(x[0]) for x in predictions]
                    ridx = 0
                    should_set_labels = False
                    labels_dict = {}
                    if "labels" in label_rules \
                       and "label_values" in label_rules:
                        label_rows = label_rules["label_values"]
                        for idx, lidx in enumerate(label_rows):
                            if len(label_rules["labels"]) >= idx:
                                should_set_labels = True
                                labels_dict[str(lidx)] = \
                                    label_rules["labels"][idx]
                    # end of compiling labels dictionary
                    log.info(("{} - ml_type={} scores={} accuracy={} "
                              "merging samples={} with predictions={} "
                              "labels={}")
                             .format(
                                label,
                                ml_type,
                                scores,
                                accuracy.get("accuracy", None),
                                len(sample_rows.index),
                                len(rounded),
                                labels_dict))
                    for idx, row in row_df.iterrows():
                        if len(sample_predictions) > max_records:
                            log.info(("{} hit max={} predictions")
                                     .format(
                                         label,
                                         max_records))
                            break
                        new_row = json.loads(row.to_json())
                        cur_value = rounded[ridx]
                        if predict_feature in row:
                            new_row["_original_{}".format(
                                    predict_feature)] = \
                                row[predict_feature]
                        else:
                            new_row["_original_{}".format(
                                    predict_feature)] = \
                                "missing-from-dataset"
                        new_row[predict_feature] = int(cur_value)
                        if should_set_labels:
                            new_row["label_name"] = \
                                labels_dict[str(int(cur_value))]
                        new_row["_row_idx"] = ridx
                        new_row["_count"] = idx
                        sample_predictions.append(new_row)
                        ridx += 1
                    # end of merging samples with predictions
                elif ml_type == "regression":
                    if new_model:
                        last_step = "building new regression model"
                        estimators = []
                        estimators.append(
                            ("standardize",
                             StandardScaler()))
                        estimators.append(
                            ("mlp",
                             model))
                        last_step = "building pipeline"
                        pipeline = Pipeline(estimators)
                        log.info(("{} - starting regression kfolds "
                                  "splits={} seed={}")
                                 .format(
                                    label,
                                    num_splits,
                                    seed))
                        last_step = "starting KFolds"
                        kfold = KFold(
                            n_splits=num_splits,
                            random_state=seed)
                        log.info(("{} - regression cross_val_score "
                                  "kfold={}")
                                 .format(
                                    label,
                                    num_splits))
                        last_step = "starting cross_val_score"
                        results = cross_val_score(
                            pipeline,
                            sample_rows.values,
                            target_rows.values,
                            cv=kfold)
                        log.info(("{} - regression prediction score: "
                                  "mean={} std={}")
                                 .format(
                                    label,
                                    results.mean(),
                                    results.std()))
                        last_step = "getting scores"
                        scores = [
                            results.std(),
                            results.mean()
                        ]
                        last_step = "getting accuracy"
                        accuracy = {
                            "accuracy": results.mean() * 100
                        }
                    # end of if new model or using existing
                    last_step = "making predictions on samples"
                    log.info(("{} - regression accuracy={} samples={}")
                             .format(
                                label,
                                accuracy["accuracy"],
                                num_samples))
                    org_predictions = model.predict(
                        sample_rows.values,
                        verbose=verbose)
                    if apply_scaler and scaler_test:
                        inverse_predictions = scaler_test.inverse_transform(
                            org_predictions.reshape(-1, 1)).reshape(-1)
                        predict_feature_values = \
                            pd.Series(inverse_predictions)
                        inverse_predictions_df = pd.DataFrame(
                            predict_feature_values,
                            columns=[predict_feature])
                        predictions = inverse_predictions_df.values
                        merge_req = {
                            "org_recs": org_df,
                            "inverse_recs": inverse_predictions_df
                        }
                        merge_res = merge_inverse_data_into_original(
                            req=merge_req,
                            sort_on_index=predict_feature,
                            ordered_columns=[predict_feature])

                        merge_df = merge_res["merge_df"]
                        sample_predictions = merge_df.to_json()
                        log.info(("{} - merge_df={}")
                                 .format(
                                    label,
                                    len(sample_predictions)))
                        are_predicts_merged = True
                    else:
                        last_step = "casting predictions to float"
                        predictions = [float(x) for x in org_predictions]

                        log.info(("{} - ml_type={} scores={} accuracy={} "
                                  "merging samples={} with predictions={}")
                                 .format(
                                    label,
                                    ml_type,
                                    scores,
                                    accuracy.get("accuracy", None),
                                    len(sample_rows.index),
                                    len(predictions)))
                        last_step = "merging predictions with org dataframe"
                        ridx = 0
                        for idx, row in row_df.iterrows():
                            if len(sample_predictions) > max_records:
                                log.info(("{} hit max={} predictions")
                                         .format(
                                            label,
                                            max_records))
                                break
                            new_row = json.loads(row.to_json())
                            cur_value = predictions[ridx]
                            if predict_feature in row:
                                new_row["_original_{}".format(
                                        predict_feature)] = \
                                    row[predict_feature]
                            else:
                                new_row["_original_{}".format(
                                        predict_feature)] = \
                                    "missing-from-dataset"
                            new_row[predict_feature] = cur_value
                            new_row["_row_idx"] = ridx
                            new_row["_count"] = idx
                            sample_predictions.append(new_row)
                            log.debug(("predicting={} target={} predicted={}")
                                      .format(
                                        predict_feature,
                                        target_rows[ridx],
                                        new_row[predict_feature]))
                            ridx += 1
                        # end of merging samples with predictions
                    # handle inverse transform for scaler datasets
                    last_step = "merging predictions done"

                else:
                    last_step = ("{} - invalid ml_type={} "
                                 "rows={}").format(
                                    label,
                                    ml_type,
                                    num_samples)
                    res["status"] = ERR
                    res["err"] = last_step
                    res["data"] = None
                    return res
            else:
                log.info(("{} - skipping predictions")
                         .format(
                             should_predict))
        except Exception as f:
            predictions = None
            last_step = ("{} - failed predicting '{}' "
                         "file={} weights={} "
                         "with ex={}").format(
                            label,
                            last_step,
                            weights_file,
                            weights_json,
                            f)
            log.error(last_step)
            res["status"] = ERR
            res["err"] = last_step
            res["data"] = None
            return res
        # end of try/ex to predict

        if ml_type == "classification":
            last_step = ("packaging {} predictions={} "
                         "rows={}").format(
                            ml_type,
                            len(rounded),
                            len(sample_rows.index))
        else:
            last_step = ("packaging {} predictions={} "
                         "rows={}").format(
                            ml_type,
                            len(sample_predictions),
                            len(sample_rows.index))

        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        try:
            image_saved = save_prediction_image(
                label=label,
                history=history,
                histories=histories,
                image_file=image_file)
            if image_saved:
                log.info(("{} - created image_file={}")
                         .format(
                             label,
                             image_file))
            else:
                image_file = None
        except Exception as f:
            image_file = None
            last_step = ("{} - failed creating image "
                         "with ex={}").format(
                            label,
                            f)
            log.error(last_step)
        # end of trying to build prediction history image

        model_weights = {}
        if new_model and save_weights:
            try:
                model_weights = {}
                # disabled for https://github.com/keras-team/keras/issues/4875
                # model.save_weights(weights_file)
                # model_weights = model.get_weights()
            except Exception as m:
                log.error(("saving label={} weights_file={} failed ex={}")
                          .format(
                            label,
                            weights_file,
                            m))
            # end of try/ex save
            log.info(("convert label={} weights_file={} to df")
                     .format(
                        label,
                        weights_file))
            try:
                log.info(("label={} weights_file={} to HDFStore")
                         .format(
                            label,
                            weights_file))
                hdf_store = \
                    pd.HDFStore(weights_file)
                log.info(("label={} HDFStore getting keys")
                         .format(
                            label))
                hdf_keys = hdf_store.keys()
                model_weights = {
                    "keys": hdf_keys,
                    "weights": []
                }
                log.info(("label={} HDFStore found weight_file={} keys={}")
                         .format(
                            label,
                            weights_file,
                            hdf_keys))
                for cur_key in hdf_keys:
                    log.info(("label={} HDFStore converting keys={} to df")
                             .format(
                                label,
                                cur_key))
                    df_hdf = \
                        pd.read_hdf(weights_file, cur_key)
                    model_weights["weights"].append({
                        "key": cur_key,
                        "df": df_hdf.to_json()})
                # end of for all keys in the HDFStore
            except Exception as m:
                log.error(("convert label={} weights_file={} failed ex={}")
                          .format(
                            label,
                            weights_file,
                            m))
            # end of try/ex convert

        # end of save_weights file for next run

        log.info(("{} - predictions done")
                 .format(
                     label))

        data["acc"] = accuracy
        data["histories"] = histories
        data["image_file"] = image_file
        if new_model:
            if ml_type == "standalone-classification":
                data["model"] = model
            else:
                data["model"] = model.model
        else:
            data["model"] = model

        data["weights"] = model_weights
        data["indexes"] = indexes
        data["scores"] = scores
        data["predictions"] = predictions
        data["rounded"] = rounded
        data["sample_predictions"] = sample_predictions
        data["confusion_matrix"] = None
        data["scaler_train"] = scaler_train
        data["scaler_test"] = scaler_test
        data["scaled_train_dataset"] = scaled_train_dataset
        data["scaled_test_dataset"] = scaled_test_dataset
        data["inverse_predictions"] = inverse_predictions
        data["apply_scaler"] = apply_scaler
        data["merge_df"] = merge_df
        data["are_predicts_merged"] = are_predicts_merged

        res["status"] = SUCCESS
        res["err"] = ""
        res["data"] = data

    except Exception as e:
        res["status"] = ERR
        if existing_model_dict:
            last_step = ("failed {} existing_model request={} hit ex={} "
                         "during last_step='{}'").format(
                            label,
                            req,
                            e,
                            last_step)
        else:
            last_step = ("failed {} request={} hit ex={} "
                         "during last_step='{}'").format(
                            label,
                            ppj(req),
                            e,
                            last_step)
        res["err"] = last_step
        res["data"] = None
        log.error(last_step)
    # end of try/ex

    return res
# end of make_predictions



import os
import json
import datetime


def ev(k, v):
    """ev

    :param k: environment variable key
    :param v: environment variable value
    """
    return os.getenv(k, v).strip().lstrip()
# end of ev


def rnow(f="%Y-%m-%d %H:%M:%S"):
    """rnow

    :param f: format for the string
    """
    return datetime.datetime.now().strftime(f)
# end of rnow


def convert_to_date(
        value=None,
        format="%Y-%m-%d %H:%M:%S"):
    """convert_to_date

    param: value - datetime object
    param: format - string format
    """

    if value:
        return value.strftime(format)

    return ""
# end of convert_to_date


def ppj(json_data):
    """ppj

    :param json_data: dictionary to print
    """
    return str(json.dumps(
                json_data,
                sort_keys=True,
                indent=4,
                separators=(',', ': ')))
# end of ppj



import glob
import copy
import random
import pandas as pd
from spylunking.log.setup_logging import build_colorized_logger
from antinex_utils.consts import VALID
from antinex_utils.consts import INVALID
from antinex_utils.utils import ev
from antinex_utils.utils import ppj
from antinex_utils.utils import rnow


log = build_colorized_logger(
    name='prepare_dataset_tools')


def find_all_headers(
        use_log_id=None,
        pipeline_files=[],
        label_rules=None):
    """find_all_headers

    :param use_log_id: label for debugging in logs
    :param pipeline_files: list of files to prep
    :param label_rules: dict of rules to apply
    """

    log_id = ""
    if use_log_id:
        log_id = use_log_id

    log.info(("{} find_all_headers - START")
             .format(
                 log_id))

    headers = ["src_file"]
    headers_dict = {"src_file": None}

    if label_rules:
        headers = ["src_file", "label_value", "label_name"]
        headers_dict = {"src_file": None,
                        "label_value": None,
                        "label_name": None}

    for c in pipeline_files:
        df = pd.read_csv(c)
        for h in df.columns.values:
            if h not in headers_dict:
                headers_dict[h] = "{}_{}".format(
                                    c,
                                    h)
                headers.append(h)
        # end for all headers in the file
    # end for all files to find common_headers

    log.info(("{} headers={}")
             .format(
                log_id,
                len(headers)))

    log.info(("{} find_all_headers - END")
             .format(
                log_id))

    return headers, headers_dict
# end of find_all_headers


def build_csv(
        pipeline_files=[],
        fulldata_file=None,
        clean_file=None,
        post_proc_rules=None,
        label_rules=None,
        use_log_id=None,
        meta_suffix="metadata.json"):
    """build_csv

    :param pipeline_files: list of files to process
    :param fulldata_file: output of non-edited merged data
    :param clean_file: cleaned csv file should be ready for training
    :param post_proc_rules: apply these rules to post processing (clean)
    :param label_rules: apply labeling rules (classification only)
    :param use_log_id: label for tracking the job in the logs
    :param meta_suffix: file suffix
    """

    save_node = {
        "status": INVALID,
        "pipeline_files": pipeline_files,
        "post_proc_rules": post_proc_rules,
        "label_rules": label_rules,
        "fulldata_file": fulldata_file,
        "fulldata_metadata_file": None,
        "clean_file": clean_file,
        "clean_metadata_file": None,
        "features_to_process": [],
        "feature_to_predict": None,
        "ignore_features": [],
        "full_headers": [],
        "clean_headers": [],
        "df_json": {},
        "version": 1
    }

    log_id = ""
    if use_log_id:
        log_id = use_log_id

    if not fulldata_file:
        log.error("missing fulldata_file - stopping")
        save_node["status"] = INVALID
        return save_node
    if not clean_file:
        log.error("missing clean_file - stopping")
        save_node["status"] = INVALID
        return save_node

    fulldata_metadata_file = "{}/fulldata_{}".format(
        "/".join(fulldata_file.split("/")[:-1]),
        meta_suffix)

    clean_metadata_file = "{}/cleaned_{}".format(
        "/".join(clean_file.split("/")[:-1]),
        meta_suffix)

    log.info(("{} build_csv - START")
             .format(
                 log_id))

    common_headers, \
        headers_dict = find_all_headers(
                            use_log_id=log_id,
                            pipeline_files=pipeline_files)

    log.info(("{} num common_headers={} headers={}")
             .format(
                log_id,
                len(common_headers),
                common_headers))

    # since the headers can be different we rebuild a new one:

    mark_default_value = None
    if "mark_empty" in post_proc_rules:
        mark_default_value = post_proc_rules["mark_empty"]
        log.info(("{} using mark_empty={}")
                 .format(
                     log_id,
                     mark_default_value))

    hdrs = {}
    for h in common_headers:
        hdrs[h] = mark_default_value

    features_to_process = []
    feature_to_predict = None
    ignore_features = []

    set_if_above = None
    labels = []
    label_values = []
    if label_rules:
        set_if_above = label_rules["set_if_above"]
        labels = label_rules["labels"]
        label_values = label_rules["label_values"]
    if post_proc_rules:
        if "predict_feature" in post_proc_rules:
            feature_to_predict = post_proc_rules["predict_feature"]
    if not feature_to_predict:
        if "label_name" in hdrs:
            feature_to_predict = "label_name"

    all_rows = []
    num_done = 1
    total_files = len(pipeline_files)
    for c in pipeline_files:
        log.info(("{} merging={}/{} csv={}")
                 .format(
                    log_id,
                    num_done,
                    total_files,
                    c))
        cf = pd.read_csv(c)
        if mark_default_value:
            log.info(("{} filling nan with value={}")
                     .format(
                        log_id,
                        mark_default_value))
            cf.fillna(value=mark_default_value,
                      inplace=True)
        # end of making sure fillna is done if requested

        log.info(("{} processing rows={}")
                 .format(
                     log_id,
                     len(cf.index)))
        for index, row in cf.iterrows():
            valid_row = True
            new_row = copy.deepcopy(hdrs)
            new_row["src_file"] = c
            for k in hdrs:
                if k in row:
                    new_row[k] = row[k]
                else:
                    if mark_default_value:
                        new_row[k] = mark_default_value
            # end of for all headers to copy in

            if label_rules:
                test_rand = random.randint(0, 100)
                if test_rand > set_if_above:
                    new_row["label_value"] = label_values[1]
                    new_row["label_name"] = labels[1]

                # if you make the "set above" greater than 100
                # it will tag the entire dataset with just 1 label
                # nice if your data is the same
                else:
                    new_row["label_value"] = label_values[0]
                    new_row["label_name"] = labels[0]
            # end of applying label rules

            if valid_row:
                all_rows.append(new_row)
        # end of for all rows in this file

        num_done += 1
    # end of building all files into one list

    log.info(("{} fulldata rows={} generating df")
             .format(
                log_id,
                len(all_rows)))

    df = pd.DataFrame(all_rows)
    log.info(("{} df rows={} headers={}")
             .format(
                log_id,
                len(df.index),
                df.columns.values))

    if ev("CONVERT_DF",
          "0") == "1":
        log.info(("{} converting df to json")
                 .format(
                    log_id))
        save_node["df_json"] = df.to_json()

    if clean_file:
        log.info(("{} writing fulldata_file={}")
                 .format(
                    log_id,
                    fulldata_file))
        df.to_csv(fulldata_file,
                  sep=',',
                  encoding='utf-8',
                  index=False)
        log.info(("{} done writing fulldata_file={}")
                 .format(
                    log_id,
                    fulldata_file))

        if post_proc_rules:

            features_to_process = []
            ignore_features = []
            if label_rules:
                if feature_to_predict:
                    ignore_features = [feature_to_predict]

            if "drop_columns" in post_proc_rules:
                for p in post_proc_rules["drop_columns"]:
                    if p in headers_dict:
                        ignore_features.append(p)
                # post proce filter more features out
                # for non-int/float types

                for d in df.columns.values:
                    add_this_one = True
                    for i in ignore_features:
                        if d == i:
                            add_this_one = False
                            break
                    if add_this_one:
                        features_to_process.append(d)
                # for all df columns we're not ignoring...
                # add them as features to process

                log.info(("{} writing fulldata metadata file={}")
                         .format(
                            log_id,
                            fulldata_metadata_file))
                header_data = {"headers": list(df.columns.values),
                               "output_type": "fulldata",
                               "pipeline_files": pipeline_files,
                               "post_proc_rules": post_proc_rules,
                               "label_rules": label_rules,
                               "features_to_process": features_to_process,
                               "feature_to_predict": feature_to_predict,
                               "ignore_features": ignore_features,
                               "created": rnow()}

                save_node["full_headers"] = list(
                    df.columns.values)
                with open(fulldata_metadata_file, "w") as otfile:
                    otfile.write(str(ppj(header_data)))

                keep_these = features_to_process
                if feature_to_predict:
                    keep_these.append(feature_to_predict)

                log.info(("{} creating new clean_file={} "
                          "keep_these={} "
                          "predict={}")
                         .format(
                            log_id,
                            clean_file,
                            keep_these,
                            feature_to_predict))

                # need to remove all columns that are all nan
                clean_df = None
                if "keep_nans" not in post_proc_rules:
                    clean_df = df[keep_these].dropna(
                                axis=1, how='all').dropna()
                else:
                    clean_df = df[keep_these].dropna(
                                axis=1, how='all')
                # allow keeping empty columns

                log.info(("{} clean_df colums={} rows={}")
                         .format(
                            log_id,
                            clean_df.columns.values,
                            len(clean_df.index)))

                if len(clean_df.columns.values) == 0:
                    log.error("Postproc clean df has no columns")
                if len(clean_df.index) == 0:
                    log.error("Postproc clean df has no rows")

                cleaned_features = clean_df.columns.values
                cleaned_to_process = []
                cleaned_ignore_features = []
                for c in cleaned_features:
                    if feature_to_predict:
                        if c == feature_to_predict:
                            cleaned_ignore_features.append(c)
                    else:
                        keep_it = True
                        for ign in ignore_features:
                            if c == ign:
                                cleaned_ignore_features.append(c)
                                keep_it = False
                                break
                        # end of for all feaures to remove
                        if keep_it:
                            cleaned_to_process.append(c)
                # end of new feature columns

                log.info(("{} writing DROPPED clean_file={} "
                          "features_to_process={} "
                          "ignore_features={} "
                          "predict={}")
                         .format(
                            log_id,
                            clean_file,
                            cleaned_to_process,
                            cleaned_ignore_features,
                            feature_to_predict))

                write_clean_df = clean_df.drop(
                    columns=cleaned_ignore_features
                )
                log.info(("cleaned_df rows={}")
                         .format(len(write_clean_df.index)))
                write_clean_df.to_csv(
                         clean_file,
                         sep=',',
                         encoding='utf-8',
                         index=False)

                clean_metadata_file = "{}/cleaned_{}".format(
                    "/".join(clean_file.split("/")[:-1]),
                    meta_suffix)
                log.info(("{} writing clean metadata file={}")
                         .format(
                            log_id,
                            clean_metadata_file))
                header_data = {"headers": list(write_clean_df.columns.values),
                               "output_type": "clean",
                               "pipeline_files": pipeline_files,
                               "post_proc_rules": post_proc_rules,
                               "label_rules": label_rules,
                               "features_to_process": cleaned_to_process,
                               "feature_to_predict": feature_to_predict,
                               "ignore_features": cleaned_ignore_features,
                               "created": rnow()}

                save_node["clean_headers"] = list(
                    write_clean_df.columns.values)
                with open(clean_metadata_file, "w") as otfile:
                    otfile.write(str(ppj(header_data)))
            else:

                for d in df.columns.values:
                    add_this_one = True
                    for i in ignore_features:
                        if d == i:
                            add_this_one = False
                            break
                    if add_this_one:
                        features_to_process.append(d)
                # for all df columns we're not ignoring...
                # add them as features to process

                log.info(("{} writing fulldata metadata file={}")
                         .format(
                            log_id,
                            fulldata_metadata_file))
                header_data = {"headers": list(df.columns.values),
                               "output_type": "fulldata",
                               "pipeline_files": pipeline_files,
                               "post_proc_rules": post_proc_rules,
                               "label_rules": label_rules,
                               "features_to_process": features_to_process,
                               "feature_to_predict": feature_to_predict,
                               "ignore_features": ignore_features,
                               "created": rnow()}

                save_node["full_headers"] = list(
                    df.columns.values)
                with open(fulldata_metadata_file, "w") as otfile:
                    otfile.write(str(ppj(header_data)))

                keep_these = features_to_process
                if feature_to_predict:
                    keep_these.append(feature_to_predict)

                log.info(("{} creating new clean_file={} "
                          "keep_these={} "
                          "predict={}")
                         .format(
                            log_id,
                            clean_file,
                            keep_these,
                            feature_to_predict))

                # need to remove all columns that are all nan
                clean_df = None
                if "keep_nans" not in post_proc_rules:
                    clean_df = df[keep_these].dropna(
                                axis=1, how='all').dropna()
                else:
                    clean_df = df[keep_these].dropna(
                                axis=1, how='all')
                # allow keeping empty columns

                log.info(("{} clean_df colums={} rows={}")
                         .format(
                            log_id,
                            clean_df.columns.values,
                            len(clean_df.index)))

                if len(clean_df.columns.values) == 0:
                    log.error(("{} The clean df has no columns")
                              .format(
                                log_id))
                if len(clean_df.index) == 0:
                    log.error(("{} The clean df has no rows")
                              .format(
                                log_id))

                cleaned_features = clean_df.columns.values
                cleaned_to_process = []
                cleaned_ignore_features = []
                for c in cleaned_features:
                    if feature_to_predict:
                        if c == feature_to_predict:
                            cleaned_ignore_features.append(c)
                    else:
                        keep_it = True
                        for ign in ignore_features:
                            if c == ign:
                                cleaned_ignore_features.append(c)
                                keep_it = False
                                break
                        # end of for all feaures to remove
                        if keep_it:
                            cleaned_to_process.append(c)
                # end of new feature columns

                log.info(("{} writing DROPPED clean_file={} "
                          "features_to_process={} "
                          "ignore_features={} "
                          "predict={}")
                         .format(
                            log_id,
                            clean_file,
                            cleaned_to_process,
                            cleaned_ignore_features,
                            feature_to_predict))

                write_clean_df = clean_df.drop(
                    columns=cleaned_ignore_features
                )
                log.info(("{} cleaned_df rows={}")
                         .format(
                            log_id,
                            len(write_clean_df.index)))
                write_clean_df.to_csv(
                         clean_file,
                         sep=',',
                         encoding='utf-8',
                         index=False)

                log.info(("{} writing clean metadata file={}")
                         .format(
                            log_id,
                            clean_metadata_file))
                header_data = {"headers": list(write_clean_df.columns.values),
                               "output_type": "clean",
                               "pipeline_files": pipeline_files,
                               "post_proc_rules": post_proc_rules,
                               "label_rules": label_rules,
                               "features_to_process": cleaned_to_process,
                               "feature_to_predict": feature_to_predict,
                               "ignore_features": cleaned_ignore_features,
                               "created": rnow()}

                save_node["clean_headers"] = list(
                    write_clean_df.columns.values)
                with open(clean_metadata_file, "w") as otfile:
                    otfile.write(str(ppj(header_data)))

            # end of if/else

            save_node["clean_file"] = clean_file
            save_node["clean_metadata_file"] = clean_metadata_file

            log.info(("{} done writing clean_file={}")
                     .format(
                        log_id,
                        clean_file))
        # end of post_proc_rules

        save_node["fulldata_file"] = fulldata_file
        save_node["fulldata_metadata_file"] = fulldata_metadata_file

        save_node["status"] = VALID
    # end of writing the file

    save_node["features_to_process"] = features_to_process
    save_node["feature_to_predict"] = feature_to_predict
    save_node["ignore_features"] = ignore_features

    log.info(("{} build_csv - END")
             .format(
                log_id))

    return save_node
# end of build_csv


def find_all_pipeline_csvs(
        use_log_id=None,
        csv_glob_path="/opt/antinex/datasets/**/*.csv"):
    """find_all_pipeline_csvs

    :param use_log_id: label for logs
    :param csv_glob_path: path to files to process
    """

    log_id = ""
    if use_log_id:
        log_id = use_log_id

    log.info(("{} finding pipeline csvs in dir={}")
             .format(
                log_id,
                csv_glob_path))

    pipeline_files = []

    for csv_file in glob.iglob(csv_glob_path,
                               recursive=True):
        log.info(("{} adding file={}")
                 .format(
                    log_id,
                    csv_file))
        pipeline_files.append(csv_file)
    # end of for all csvs

    log.info(("{} pipeline files={}")
             .format(
                log_id,
                len(pipeline_files)))

    return pipeline_files
# end of find_all_pipeline_csvs



import json
import pandas as pd
from spylunking.log.setup_logging import build_colorized_logger
from antinex_utils.consts import SUCCESS
from antinex_utils.consts import ERR
from antinex_utils.consts import NOTRUN


log = build_colorized_logger(
    name='merge_inverse_data_into_original')


def merge_inverse_data_into_original(
        req,
        sort_on_index=None,
        ordered_columns=None):
    """merge_inverse_data_into_original

    :param req: managed dictionary
    :param sort_on_index: sort the dataframe on this column name
    :param ordered_columns: column list to rename the inverse transform
    """
    label = req.get(
        "label",
        "")
    last_step = "not-run"
    status = NOTRUN
    org_df = None
    predict_df = None
    merge_df = None
    res = {
        "status": status,
        "err": last_step,
        "sorted_org_df": org_df,
        "predict_df": predict_df,
        "merge_df": merge_df
    }
    try:

        last_step = ("sorting org_records on index={}").format(
                        sort_on_index)
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        org_df = pd.DataFrame(req["org_recs"]).set_index(sort_on_index)
        inverse_df = pd.DataFrame(req["inverse_recs"])
        if ordered_columns:
            last_step = "getting inverse column ordering={}".format(
                ordered_columns)
            log.info(("{} - {}")
                     .format(
                        label,
                        last_step))
            inverse_df.columns = ordered_columns
        if sort_on_index:
            last_step = ("sorting inverse on index={}").format(
                            sort_on_index)
            log.info(("{} - {}")
                     .format(
                        label,
                        last_step))
            inverse_df.set_index(sort_on_index)

        row_idx = 0
        merge_rows = []
        predict_rows = []
        last_step = ("for org_value rows={}").format(
                        len(org_df.index))
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        predicted_index = list(org_df.index)
        for idx, row in org_df.iterrows():
            inv_row = inverse_df.iloc[row_idx].to_json()
            inv_dict = json.loads(inv_row)
            predict_rows.append(inv_dict)
            cur_dict = json.loads(row.to_json())
            cur_dict[sort_on_index] = predicted_index[row_idx]
            cur_dict["_row_idx"] = row_idx
            for i in inv_dict:
                if i in cur_dict:
                    cur_dict["_predicted_{}".format(
                       i)] = inv_dict[i]
            merge_rows.append(cur_dict)
            row_idx += 1
        # end of for all rows to check for ordering

        last_step = ("building predict_df rows={}").format(
                        len(predict_rows))
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        predict_df = pd.DataFrame(
            predict_rows)
        merge_df = pd.DataFrame(
            merge_rows)

        if sort_on_index:
            last_step = ("sorting={} predict_df rows={}").format(
                            sort_on_index,
                            len(predict_rows))
            log.info(("{} - {}")
                     .format(
                        label,
                        last_step))
            predict_df.set_index(sort_on_index)
            merge_df.set_index(sort_on_index)

        last_step = ""
        status = SUCCESS

    except Exception as e:
        last_step = ("failed merge_inverse_data_into_original "
                     "with ex={} last_step='{}' "
                     "org_recs={} inverse_recs={}").format(
                        e,
                        last_step,
                        len(org_df.index),
                        len(inverse_df.index))
        log.error(("{} - {}")
                  .format(
                    label,
                    last_step))
        status = ERR
    # end of try/ex

    res = {
        "status": status,
        "err": last_step,
        "sorted_org_df": org_df,
        "predict_df": predict_df,
        "merge_df": merge_df
    }

    return res
# end of merge_inverse_data_into_original



import os
import sys
import warnings
import unittest

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from setuptools.command.test import test as TestCommand

"""
https://packaging.python.org/guides/making-a-pypi-friendly-readme/
check the README.rst works on pypi as the
long_description with:
twine check dist/*
"""
long_description = open('README.rst').read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


long_description = open('README.rst').read()

cur_path, cur_script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(cur_path))

requires_that_fail_on_rtd = [
    "h5py",
    "keras",
    "tables",
    "tensorflow"
]

install_requires = [
    "celery>=4.1.0",
    "celery-connectors",
    "celery-loaders",
    "colorlog",
    "coverage",
    "flake8",
    "future",
    "matplotlib",
    "numpy",
    "pandas",
    "pep8",
    "pipenv",
    "pycodestyle",
    "pydocstyle",
    "pylint",
    "recommonmark",
    "requests",
    "scikit-learn",
    "sphinx",
    "sphinx-autobuild",
    "sphinx_rtd_theme",
    "spylunking",
    "tox",
    "unittest2",
    "mock"
]

# if not on readthedocs.io get all the pips:
if os.getenv("READTHEDOCS", "") == "":
    install_requires = install_requires + requires_that_fail_on_rtd

if sys.version_info < (3, 5):
    warnings.warn(
        "Less than Python 3.5 is not supported.",
        DeprecationWarning)


# Do not import antinex_utils module here, since deps may not be installed
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "antinex_utils"))

setup(
    name="antinex-utils",
    cmdclass={"test": PyTest},
    version="1.2.4",
    description="AntiNex Utilities for Keras and Tensorflow",
    long_description_content_type='text/x-rst',
    long_description=long_description,
    author="Jay Johnson",
    author_email="jay.p.h.johnson@gmail.com",
    url="https://github.com/jay-johnson/antinex-utils",
    packages=[
        "antinex_utils",
        "antinex_utils.log"
    ],
    package_data={},
    install_requires=install_requires,
    test_suite="setup.antinex_utils_test_suite",
    tests_require=[
        "pytest"
    ],
    scripts=[
    ],
    use_2to3=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ])



import pandas as pd
from spylunking.log.setup_logging import build_colorized_logger
from antinex_utils.consts import SUCCESS
from antinex_utils.consts import ERR
from antinex_utils.consts import FAILED
from antinex_utils.consts import NOTRUN
from antinex_utils.build_scaler_dataset_from_records import \
    build_scaler_dataset_from_records
from sklearn.model_selection import train_test_split


log = build_colorized_logger(
    name='build_scaler_train_and_test_datasets')


def build_scaler_train_and_test_datasets(
        label,
        train_features,
        test_feature,
        df,
        test_size,
        seed,
        scaler_cast_to_type="float32",
        min_feature_range=-1,
        max_feature_range=1):
    """build_scaler_train_and_test_datasets

    :param label: log label
    :param train_features: features to train
    :param test_feature: target feature name
    :param df: dataframe to build scalers and test and train datasets
    :param test_size: percent of test to train rows
    :param min_feature_range: min scaler range
    :param max_feature_range: max scaler range
    """

    status = NOTRUN
    last_step = "not-run"
    scaled_train_df = None
    scaled_test_df = None
    scaled_train_dataset = None
    scaled_test_dataset = None
    scaler_train = None
    scaler_test = None
    x_train = None
    y_train = None
    x_test = None
    y_test = None
    data = {
        "scaled_train_df": scaled_train_df,
        "scaled_test_df": scaled_test_df,
        "scaled_train_dataset": scaled_train_dataset,
        "scaled_test_dataset": scaled_test_dataset,
        "scaler_train": scaler_train,
        "scaler_test": scaler_test,
        "x_train": x_train,
        "y_train": y_train,
        "x_test": x_test,
        "y_test": y_test,
        "min_range": min_feature_range,
        "max_range": max_feature_range
    }
    res = {
        "status": status,
        "err": last_step,
        "data": data
    }
    try:
        last_step = ("building scalers df.rows={} columns={} "
                     "train_features={} test_feature={}").format(
                        len(df.index),
                        list(df.columns.values),
                        train_features,
                        test_feature)
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        scaled_train_df = df[train_features]
        scaled_test_df = pd.DataFrame(
            {test_feature: df[test_feature]})

        last_step = ("building scaled TRAIN dataset [{},{}]").format(
                    min_feature_range,
                    max_feature_range)
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        scaled_train_res = \
            build_scaler_dataset_from_records(
                label=label,
                record_list=scaled_train_df.to_json(),
                min_feature=min_feature_range,
                max_feature=max_feature_range,
                cast_to_type=scaler_cast_to_type)

        last_step = ("building scaled TEST dataset [{},{}]").format(
                        min_feature_range,
                        max_feature_range)
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))
        scaled_test_res = \
            build_scaler_dataset_from_records(
                label=label,
                record_list=scaled_test_df.to_json(),
                min_feature=min_feature_range,
                max_feature=max_feature_range,
                cast_to_type=scaler_cast_to_type)

        last_step = ("scaled dataset transform "
                     "train_status={} test_status={}").format(
                        scaled_train_res["status"] == SUCCESS,
                        scaled_test_res["status"] == SUCCESS)

        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))

        if scaled_train_res["status"] == SUCCESS \
           and scaled_test_res["status"] == SUCCESS:
            last_step = ("scaled train_rows={} "
                         "test_rows={}").format(
                            len(scaled_train_res["dataset"]),
                            len(scaled_test_res["dataset"]))
            log.info(("{} - {}")
                     .format(
                        label,
                        last_step))
            scaler_train = scaled_train_res["scaler"]
            scaler_test = scaled_test_res["scaler"]
            scaled_train_dataset = scaled_train_res["dataset"]
            scaled_test_dataset = scaled_test_res["dataset"]
            (x_train,
             x_test,
             y_train,
             y_test) = train_test_split(
                scaled_train_dataset,
                scaled_test_dataset,
                test_size=test_size,
                random_state=seed)
        else:
            last_step = ("failed dataset transform "
                         "train_status={} test_status={}").format(
                             scaled_train_res["status"],
                             scaled_test_res["status"])
            log.error(("{} - {}")
                      .format(
                        label,
                        last_step))

            status = FAILED
            res = {
                "status": status,
                "err": last_step,
                "data": data
            }
            return res
        # if built both train and test successfully

        last_step = ("train_rows={} test_rows={} "
                     "x_train={} x_test={} "
                     "y_train={} y_test={}").format(
                        len(scaled_train_df.index),
                        len(scaled_test_df),
                        len(scaled_train_df.index),
                        len(scaled_test_df.index),
                        len(scaled_train_dataset),
                        len(scaled_test_dataset))
        log.info(("{} - {}")
                 .format(
                    label,
                    last_step))

        data["scaled_train_df"] = scaled_train_df
        data["scaled_test_df"] = scaled_test_df
        data["scaled_train_dataset"] = scaled_train_dataset
        data["scaled_test_dataset"] = scaled_test_dataset
        data["scaler_train"] = scaler_train
        data["scaler_test"] = scaler_test
        data["x_train"] = x_train
        data["y_train"] = y_train
        data["x_test"] = x_test
        data["y_test"] = y_test
        status = SUCCESS
        last_step = ""

        log.info(("{} - done")
                 .format(
                    label))
    except Exception as e:
        last_step = ("failed during last_step='{}' with ex={} "
                     "building scalers df.rows={} columns={} "
                     "train_features={} test_feature={}").format(
                        last_step,
                        e,
                        len(df.index),
                        list(df.columns.values),
                        train_features,
                        test_feature)
        log.error(("{} - {}")
                  .format(
                    label,
                    last_step))
        status = ERR
    # if applying scaler to predict rows

    res = {
        "status": status,
        "err": last_step,
        "data": data
    }

    return res
# end of build_scaler_train_and_test_datasets
